"""
PVfit testing: Single-diode equation (SDE) inference.

Copyright 2023 Intelligent Measurement Systems LLC
"""

import importlib.resources
import json
import os
from pathlib import Path

import numpy
import pandas
import pytest
import scipy

import pvfit
from pvfit.measurement.iv.types import IVCurve
import pvfit.modeling.dc.single_diode.equation.simple.inference_iv_curve as inference_iv_curve
from pvfit.modeling.dc.single_diode.equation.simple.types import (
    ModelParametersFittableFixedProvided,
    ModelParametersFittableProvided,
    ModelParametersUnfittable,
)

ARTIFACTS_PARENT_DIR_PATH = Path(os.path.dirname(os.path.dirname(pvfit.__file__)))

if ARTIFACTS_PARENT_DIR_PATH.name == "site-packages":
    # Covers use case where one is not developing locally or running in CI.
    TEST_ARTIFACTS_PATH = Path().resolve()  # Current working directory.
else:
    TEST_ARTIFACTS_PATH = ARTIFACTS_PARENT_DIR_PATH.joinpath(
        "artifacts",
        "test",
        "sde",
        "inference",
    )

TEST_ARTIFACTS_PATH.mkdir(parents=True, exist_ok=True)
IVCURVES_MULTIPLEXED_PATH = importlib.resources.files("ivcurves")


@pytest.fixture(
    ids=["case1", "case2", "case3a", "case3b", "case3c", "case3d"],
    params=[
        {
            "given": {
                "test_set": "case1",
                "rtol": 1e-05,
                "atol": 1e-08,
            },
            "expected": {
                "model_parameters": [
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0,
                        "I_rs_A": 5.000000000000269e-10,
                        "n": 1.0100000000000022,
                        "R_s_Ohm": 0.09999999999998008,
                        "G_p_S": 0.003333333333333333,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0,
                        "I_rs_A": 4.99999999999996e-10,
                        "n": 1.2999999999999994,
                        "R_s_Ohm": 0.10000000000000976,
                        "G_p_S": 0.0033333333333333327,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0,
                        "I_rs_A": 4.999999999999962e-10,
                        "n": 1.0099999999999996,
                        "R_s_Ohm": 0.10000000000000611,
                        "G_p_S": 0.000333333333333332,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0,
                        "I_rs_A": 4.999999999999983e-10,
                        "n": 1.2999999999999996,
                        "R_s_Ohm": 0.09999999999999591,
                        "G_p_S": 0.00033333333333333305,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9999999999999999,
                        "I_rs_A": 5.000000000000212e-10,
                        "n": 1.0100000000000018,
                        "R_s_Ohm": 0.9999999999999877,
                        "G_p_S": 0.003333333333333333,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0,
                        "I_rs_A": 5.000000000000057e-10,
                        "n": 1.3000000000000005,
                        "R_s_Ohm": 0.9999999999999962,
                        "G_p_S": 0.003333333333333334,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0,
                        "I_rs_A": 4.999999999999842e-10,
                        "n": 1.0099999999999985,
                        "R_s_Ohm": 1.0000000000000075,
                        "G_p_S": 0.000333333333333335,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9999999999999999,
                        "I_rs_A": 4.999999999999922e-10,
                        "n": 1.2999999999999987,
                        "R_s_Ohm": 1.0000000000000073,
                        "G_p_S": 0.0003333333333333328,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0000000000000002,
                        "I_rs_A": 3.0000000000000024e-08,
                        "n": 1.01,
                        "R_s_Ohm": 0.10000000000000078,
                        "G_p_S": 0.003333333333333334,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0,
                        "I_rs_A": 3.00000000000003e-08,
                        "n": 1.3000000000000005,
                        "R_s_Ohm": 0.09999999999999007,
                        "G_p_S": 0.0033333333333333314,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0,
                        "I_rs_A": 2.9999999999999825e-08,
                        "n": 1.0099999999999996,
                        "R_s_Ohm": 0.10000000000000676,
                        "G_p_S": 0.0003333333333333324,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0,
                        "I_rs_A": 2.999999999999981e-08,
                        "n": 1.2999999999999994,
                        "R_s_Ohm": 0.10000000000000477,
                        "G_p_S": 0.00033333333333333495,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0,
                        "I_rs_A": 3.000000000000056e-08,
                        "n": 1.010000000000001,
                        "R_s_Ohm": 0.9999999999999948,
                        "G_p_S": 0.0033333333333333327,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0,
                        "I_rs_A": 2.9999999999999494e-08,
                        "n": 1.2999999999999985,
                        "R_s_Ohm": 1.0000000000000062,
                        "G_p_S": 0.0033333333333333353,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0,
                        "I_rs_A": 2.999999999999973e-08,
                        "n": 1.0099999999999993,
                        "R_s_Ohm": 1.000000000000002,
                        "G_p_S": 0.0003333333333333342,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0000000000000002,
                        "I_rs_A": 3.000000000000002e-08,
                        "n": 1.2999999999999998,
                        "R_s_Ohm": 0.9999999999999987,
                        "G_p_S": 0.00033333333333333446,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.999999999999999,
                        "I_rs_A": 5.000000000000075e-10,
                        "n": 1.0100000000000007,
                        "R_s_Ohm": 0.09999999999999935,
                        "G_p_S": 0.003333333333333325,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0,
                        "I_rs_A": 5.000000000000057e-10,
                        "n": 1.3000000000000007,
                        "R_s_Ohm": 0.09999999999999996,
                        "G_p_S": 0.0033333333333333296,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.999999999999999,
                        "I_rs_A": 5.000000000000016e-10,
                        "n": 1.01,
                        "R_s_Ohm": 0.09999999999999966,
                        "G_p_S": 0.00033333333333333164,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0,
                        "I_rs_A": 5.000000000000265e-10,
                        "n": 1.300000000000003,
                        "R_s_Ohm": 0.09999999999999883,
                        "G_p_S": 0.00033333333333333126,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0,
                        "I_rs_A": 5.000000000000062e-10,
                        "n": 1.0100000000000005,
                        "R_s_Ohm": 0.9999999999999996,
                        "G_p_S": 0.003333333333333332,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0,
                        "I_rs_A": 4.999999999999938e-10,
                        "n": 1.2999999999999994,
                        "R_s_Ohm": 1.0000000000000004,
                        "G_p_S": 0.0033333333333333357,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0,
                        "I_rs_A": 5.000000000000121e-10,
                        "n": 1.010000000000001,
                        "R_s_Ohm": 0.9999999999999994,
                        "G_p_S": 0.0003333333333333223,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0,
                        "I_rs_A": 5.000000000000115e-10,
                        "n": 1.3000000000000012,
                        "R_s_Ohm": 0.9999999999999993,
                        "G_p_S": 0.00033333333333332384,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0,
                        "I_rs_A": 3.000000000000002e-08,
                        "n": 1.0099999999999998,
                        "R_s_Ohm": 0.09999999999999978,
                        "G_p_S": 0.0033333333333333296,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0,
                        "I_rs_A": 3.000000000000062e-08,
                        "n": 1.3000000000000012,
                        "R_s_Ohm": 0.09999999999999865,
                        "G_p_S": 0.0033333333333333192,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0,
                        "I_rs_A": 2.999999999999951e-08,
                        "n": 1.0099999999999991,
                        "R_s_Ohm": 0.10000000000000069,
                        "G_p_S": 0.0003333333333333319,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0,
                        "I_rs_A": 3.0000000000000765e-08,
                        "n": 1.3000000000000016,
                        "R_s_Ohm": 0.09999999999999895,
                        "G_p_S": 0.00033333333333332215,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.999999999999999,
                        "I_rs_A": 3e-08,
                        "n": 1.0099999999999998,
                        "R_s_Ohm": 0.9999999999999999,
                        "G_p_S": 0.003333333333333334,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0,
                        "I_rs_A": 2.9999999999999534e-08,
                        "n": 1.299999999999999,
                        "R_s_Ohm": 1.0000000000000004,
                        "G_p_S": 0.003333333333333345,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.999999999999999,
                        "I_rs_A": 3.00000000000008e-08,
                        "n": 1.0100000000000013,
                        "R_s_Ohm": 0.9999999999999991,
                        "G_p_S": 0.00033333333333331993,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0,
                        "I_rs_A": 2.999999999999989e-08,
                        "n": 1.2999999999999996,
                        "R_s_Ohm": 1.0,
                        "G_p_S": 0.0003333333333333417,
                    },
                ],
            },
        },
        {
            "given": {
                "test_set": "case2",
                "rtol": 1e-05,
                "atol": 1e-08,
            },
            "expected": {
                "model_parameters": [
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49999999999999994,
                        "I_rs_A": 1.0000000000000201e-09,
                        "n": 1.3000000000000012,
                        "R_s_Ohm": 0.0999999999999438,
                        "G_p_S": 0.003333333333333333,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49999999999999983,
                        "I_rs_A": 1.0000000000000427e-09,
                        "n": 1.500000000000003,
                        "R_s_Ohm": 0.09999999999988081,
                        "G_p_S": 0.0033333333333333322,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5,
                        "I_rs_A": 1.0000000000000032e-09,
                        "n": 1.3000000000000003,
                        "R_s_Ohm": 0.1000000000000165,
                        "G_p_S": 0.00033333333333333354,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5,
                        "I_rs_A": 9.999999999999238e-10,
                        "n": 1.4999999999999942,
                        "R_s_Ohm": 0.10000000000007547,
                        "G_p_S": 0.00033333333333333457,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.4999999999999998,
                        "I_rs_A": 1.0000000000000443e-09,
                        "n": 1.300000000000003,
                        "R_s_Ohm": 0.9999999999999182,
                        "G_p_S": 0.003333333333333332,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49999999999999917,
                        "I_rs_A": 1.000000000000119e-09,
                        "n": 1.5000000000000093,
                        "R_s_Ohm": 0.9999999999995441,
                        "G_p_S": 0.0033333333333333275,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5,
                        "I_rs_A": 9.999999999999916e-10,
                        "n": 1.2999999999999996,
                        "R_s_Ohm": 1.0000000000000238,
                        "G_p_S": 0.0003333333333333335,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5,
                        "I_rs_A": 1.0000000000000135e-09,
                        "n": 1.5000000000000009,
                        "R_s_Ohm": 0.9999999999999816,
                        "G_p_S": 0.0003333333333333336,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5000000000000002,
                        "I_rs_A": 9.999999999999787e-09,
                        "n": 1.2999999999999985,
                        "R_s_Ohm": 0.10000000000007676,
                        "G_p_S": 0.003333333333333335,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5000000000000001,
                        "I_rs_A": 1.0000000000000169e-08,
                        "n": 1.5000000000000016,
                        "R_s_Ohm": 0.10000000000001652,
                        "G_p_S": 0.003333333333333333,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5,
                        "I_rs_A": 1.0000000000000245e-08,
                        "n": 1.3000000000000018,
                        "R_s_Ohm": 0.09999999999996696,
                        "G_p_S": 0.0003333333333333329,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5,
                        "I_rs_A": 9.99999999999972e-09,
                        "n": 1.4999999999999978,
                        "R_s_Ohm": 0.10000000000004394,
                        "G_p_S": 0.00033333333333333365,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5000000000000002,
                        "I_rs_A": 9.999999999999403e-09,
                        "n": 1.2999999999999954,
                        "R_s_Ohm": 1.000000000000137,
                        "G_p_S": 0.0033333333333333353,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5000000000000002,
                        "I_rs_A": 9.999999999999512e-09,
                        "n": 1.4999999999999956,
                        "R_s_Ohm": 1.0000000000001181,
                        "G_p_S": 0.003333333333333335,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5000000000000001,
                        "I_rs_A": 1.0000000000000338e-08,
                        "n": 1.3000000000000025,
                        "R_s_Ohm": 0.9999999999999375,
                        "G_p_S": 0.00033333333333333305,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5,
                        "I_rs_A": 1.0000000000000252e-08,
                        "n": 1.5000000000000022,
                        "R_s_Ohm": 0.9999999999999826,
                        "G_p_S": 0.00033333333333333245,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4999999999999996,
                        "I_rs_A": 1.000000000000005e-09,
                        "n": 1.3000000000000003,
                        "R_s_Ohm": 0.09999999999999278,
                        "G_p_S": 0.0033333333333333335,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5,
                        "I_rs_A": 9.99999999999978e-10,
                        "n": 1.4999999999999984,
                        "R_s_Ohm": 0.10000000000000225,
                        "G_p_S": 0.003333333333333335,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5,
                        "I_rs_A": 9.99999999999995e-10,
                        "n": 1.2999999999999994,
                        "R_s_Ohm": 0.09999999999999584,
                        "G_p_S": 0.0003333333333333326,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5,
                        "I_rs_A": 9.999999999999868e-10,
                        "n": 1.4999999999999991,
                        "R_s_Ohm": 0.10000000000000436,
                        "G_p_S": 0.0003333333333333334,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5,
                        "I_rs_A": 9.999999999999897e-10,
                        "n": 1.2999999999999994,
                        "R_s_Ohm": 1.0000000000000027,
                        "G_p_S": 0.0033333333333333327,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5000000000000004,
                        "I_rs_A": 9.999999999999736e-10,
                        "n": 1.4999999999999982,
                        "R_s_Ohm": 1.0000000000000109,
                        "G_p_S": 0.0033333333333333322,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5,
                        "I_rs_A": 1.0000000000000092e-09,
                        "n": 1.3000000000000003,
                        "R_s_Ohm": 0.9999999999999976,
                        "G_p_S": 0.00033333333333333365,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5,
                        "I_rs_A": 1.0000000000000327e-09,
                        "n": 1.5000000000000022,
                        "R_s_Ohm": 0.9999999999999897,
                        "G_p_S": 0.0003333333333333318,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4999999999999996,
                        "I_rs_A": 9.999999999999959e-09,
                        "n": 1.2999999999999996,
                        "R_s_Ohm": 0.10000000000000453,
                        "G_p_S": 0.00333333333333333,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5,
                        "I_rs_A": 1.000000000000013e-08,
                        "n": 1.5000000000000009,
                        "R_s_Ohm": 0.09999999999999926,
                        "G_p_S": 0.003333333333333331,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5,
                        "I_rs_A": 1.0000000000000281e-08,
                        "n": 1.3000000000000018,
                        "R_s_Ohm": 0.09999999999999326,
                        "G_p_S": 0.000333333333333333,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5,
                        "I_rs_A": 1.0000000000000404e-08,
                        "n": 1.500000000000003,
                        "R_s_Ohm": 0.09999999999998553,
                        "G_p_S": 0.00033333333333333115,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5,
                        "I_rs_A": 9.999999999999587e-09,
                        "n": 1.2999999999999972,
                        "R_s_Ohm": 1.0000000000000113,
                        "G_p_S": 0.0033333333333333366,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5000000000000004,
                        "I_rs_A": 9.999999999999692e-09,
                        "n": 1.4999999999999976,
                        "R_s_Ohm": 1.0000000000000113,
                        "G_p_S": 0.003333333333333335,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5,
                        "I_rs_A": 9.999999999999985e-09,
                        "n": 1.2999999999999998,
                        "R_s_Ohm": 1.0,
                        "G_p_S": 0.0003333333333333337,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5,
                        "I_rs_A": 9.999999999999831e-09,
                        "n": 1.4999999999999987,
                        "R_s_Ohm": 1.0000000000000056,
                        "G_p_S": 0.000333333333333333,
                    },
                ],
            },
        },
        {
            "given": {
                "test_set": "case3a",
                "rtol": 1e-05,
                "atol": 1e-08,
            },
            "expected": {
                "model_parameters": [
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.00156581121952,
                        "I_rs_A": 9.052254203691839e-10,
                        "n": 1.0353938169997223,
                        "R_s_Ohm": 0.08143912443503845,
                        "G_p_S": 0.00021102504883210545,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.012208523457033,
                        "I_rs_A": 5.409107659018644e-10,
                        "n": 1.013413634867728,
                        "R_s_Ohm": 0.0963238201247794,
                        "G_p_S": 0.0009763575243971334,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.994003790720126,
                        "I_rs_A": 3.216696650695513e-10,
                        "n": 0.9915524899173723,
                        "R_s_Ohm": 0.10964636191987445,
                        "G_p_S": 0.0003136187433082198,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.996571860344162,
                        "I_rs_A": 2.1202539912806434e-10,
                        "n": 0.9749341467268949,
                        "R_s_Ohm": 0.1210109969906554,
                        "G_p_S": 0.0,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.001974034021742,
                        "I_rs_A": 4.150561095678614e-10,
                        "n": 1.0020641704005773,
                        "R_s_Ohm": 0.10434138881573936,
                        "G_p_S": 0.00016951259421550685,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.008186036049807,
                        "I_rs_A": 1.0084357314911116e-09,
                        "n": 1.040739035992204,
                        "R_s_Ohm": 0.08362210452995332,
                        "G_p_S": 0.00035286216782681723,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.009401507495461,
                        "I_rs_A": 4.5418471808663815e-10,
                        "n": 1.0058766160265868,
                        "R_s_Ohm": 0.10271517902448458,
                        "G_p_S": 0.00041171932194224457,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.999791884787038,
                        "I_rs_A": 2.801615667789003e-10,
                        "n": 0.9860732689352895,
                        "R_s_Ohm": 0.11236511982172667,
                        "G_p_S": 0.0005800743081171647,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.00569010193704,
                        "I_rs_A": 2.9243954154956235e-10,
                        "n": 0.9877622784530378,
                        "R_s_Ohm": 0.11185591901756729,
                        "G_p_S": 0.00028610659161859,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.01958153581084,
                        "I_rs_A": 7.610497686777689e-10,
                        "n": 1.0280732470453868,
                        "R_s_Ohm": 0.08749338737997499,
                        "G_p_S": 0.0003971701470071563,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.009587589637254,
                        "I_rs_A": 5.265423432414597e-10,
                        "n": 1.0118241640212875,
                        "R_s_Ohm": 0.09642248070229303,
                        "G_p_S": 0.0006276651611895418,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.997154170843948,
                        "I_rs_A": 5.817476113371505e-10,
                        "n": 1.0163608694358428,
                        "R_s_Ohm": 0.09559811021746759,
                        "G_p_S": 0.0003648645060268345,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.999366831022037,
                        "I_rs_A": 2.56782849062014e-10,
                        "n": 0.9824426414628596,
                        "R_s_Ohm": 0.11495820112910783,
                        "G_p_S": 0.00026058018803942995,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.009623261274621,
                        "I_rs_A": 4.983589773410359e-10,
                        "n": 1.009684584948523,
                        "R_s_Ohm": 0.09781009603838833,
                        "G_p_S": 0.0007295762582308793,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.008277825608857,
                        "I_rs_A": 7.969471289251435e-10,
                        "n": 1.030284802673514,
                        "R_s_Ohm": 0.08904759474544477,
                        "G_p_S": 0.0008657623151434602,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.99836165131404,
                        "I_rs_A": 4.5824309207695624e-10,
                        "n": 1.0063474412568612,
                        "R_s_Ohm": 0.10388528872921565,
                        "G_p_S": 1.8640200299819246e-05,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.995496122539305,
                        "I_rs_A": 7.168791613937709e-10,
                        "n": 1.0254880344429371,
                        "R_s_Ohm": 0.09013314473407105,
                        "G_p_S": 0.00018881919384733126,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.0057424374997,
                        "I_rs_A": 6.620533206869249e-10,
                        "n": 1.0220319326933966,
                        "R_s_Ohm": 0.09312878922216931,
                        "G_p_S": 0.00036078657153186907,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.005711997691751,
                        "I_rs_A": 4.4226245059373307e-10,
                        "n": 1.0046568562032472,
                        "R_s_Ohm": 0.1013513267618965,
                        "G_p_S": 0.0006864694604337675,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.998504287226888,
                        "I_rs_A": 7.673432989118436e-10,
                        "n": 1.028102482717502,
                        "R_s_Ohm": 0.08691058519937983,
                        "G_p_S": 0.0003297870971256424,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.987676300858112,
                        "I_rs_A": 1.678767431490843e-10,
                        "n": 0.9656207829057492,
                        "R_s_Ohm": 0.12350740966680596,
                        "G_p_S": 0.00013397553042344458,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.001464606102992,
                        "I_rs_A": 5.790341097610332e-10,
                        "n": 1.0164669347288584,
                        "R_s_Ohm": 0.09859716904537072,
                        "G_p_S": 3.946706785138822e-05,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.00435973402833,
                        "I_rs_A": 9.015625694255479e-10,
                        "n": 1.0360652440810159,
                        "R_s_Ohm": 0.08985653378736677,
                        "G_p_S": 0.00036604831640288735,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.996699208397625,
                        "I_rs_A": 4.045030056392567e-10,
                        "n": 1.0013837590014256,
                        "R_s_Ohm": 0.10858497549891681,
                        "G_p_S": 0.0001701244370036583,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.000957098380404,
                        "I_rs_A": 2.9148347998220885e-10,
                        "n": 0.9876332098272315,
                        "R_s_Ohm": 0.11366847539504187,
                        "G_p_S": 0.0003100717289616421,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.999753151076565,
                        "I_rs_A": 2.6276803717365947e-10,
                        "n": 0.9836929066460628,
                        "R_s_Ohm": 0.11476580686309079,
                        "G_p_S": 0.0006256052921342581,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.999085725887511,
                        "I_rs_A": 2.734287961277672e-10,
                        "n": 0.9848091689708071,
                        "R_s_Ohm": 0.11076884985770048,
                        "G_p_S": 0.00048158352647508125,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.001116079087753,
                        "I_rs_A": 8.400946715966017e-10,
                        "n": 1.032812033706127,
                        "R_s_Ohm": 0.09053754793856997,
                        "G_p_S": 0.0002734821021955908,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.998342190868724,
                        "I_rs_A": 3.8635293828437486e-10,
                        "n": 0.9989699619547855,
                        "R_s_Ohm": 0.10643040795792726,
                        "G_p_S": 0.0,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.995881356997555,
                        "I_rs_A": 5.955919444507345e-10,
                        "n": 1.0175167224673258,
                        "R_s_Ohm": 0.0953006870370625,
                        "G_p_S": 0.0002449814166108493,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.000657105337076,
                        "I_rs_A": 2.1282803385987715e-10,
                        "n": 0.9749099475030208,
                        "R_s_Ohm": 0.11827481859445185,
                        "G_p_S": 0.0004964403036700842,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.981758087913774,
                        "I_rs_A": 4.398827640944435e-10,
                        "n": 1.0046154563381693,
                        "R_s_Ohm": 0.10135832612987693,
                        "G_p_S": 0.00031354559270175495,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.99250356930967,
                        "I_rs_A": 5.143435921976739e-10,
                        "n": 1.0112294005835973,
                        "R_s_Ohm": 0.10132640787282293,
                        "G_p_S": 0.00022374437818060524,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.003357233361745,
                        "I_rs_A": 5.394183223032944e-10,
                        "n": 1.013572267278007,
                        "R_s_Ohm": 0.10068131425996535,
                        "G_p_S": 0.0003797384258801089,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.999674400854034,
                        "I_rs_A": 7.535567416922522e-10,
                        "n": 1.0275986103443175,
                        "R_s_Ohm": 0.08830500971001216,
                        "G_p_S": 0.00037092628008957934,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.00375996097543,
                        "I_rs_A": 2.475946668103389e-10,
                        "n": 0.9811336974815716,
                        "R_s_Ohm": 0.11570692051977835,
                        "G_p_S": 0.0008587909563718019,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.989681160116382,
                        "I_rs_A": 2.0119734772182173e-10,
                        "n": 0.972828779455203,
                        "R_s_Ohm": 0.11900667854778002,
                        "G_p_S": 0.0003912011889337581,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.01489398991812,
                        "I_rs_A": 7.020811449870305e-10,
                        "n": 1.0247258635609104,
                        "R_s_Ohm": 0.09329459271825546,
                        "G_p_S": 0.0003396241313673714,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.018779718288318,
                        "I_rs_A": 4.0644366222804387e-10,
                        "n": 1.0009666378387634,
                        "R_s_Ohm": 0.10097076093566218,
                        "G_p_S": 0.0006965017224692362,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.998710671186315,
                        "I_rs_A": 8.703263985770277e-10,
                        "n": 1.0338216827305395,
                        "R_s_Ohm": 0.08368432602895853,
                        "G_p_S": 0.0,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.991085848437366,
                        "I_rs_A": 3.5038118359305086e-10,
                        "n": 0.9952130840441233,
                        "R_s_Ohm": 0.11009125047872158,
                        "G_p_S": 0.0,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.003409343972235,
                        "I_rs_A": 6.341741303133576e-10,
                        "n": 1.0199613898413293,
                        "R_s_Ohm": 0.09296189553950084,
                        "G_p_S": 0.00024039343111820652,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.001959419482143,
                        "I_rs_A": 3.7815111409984576e-10,
                        "n": 0.998046979013115,
                        "R_s_Ohm": 0.10210095707337886,
                        "G_p_S": 0.0006449634171888937,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.002609657716441,
                        "I_rs_A": 6.407258889831509e-10,
                        "n": 1.020534094916965,
                        "R_s_Ohm": 0.0943065578857104,
                        "G_p_S": 0.00031946239374595436,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.991132359785082,
                        "I_rs_A": 4.748393575062793e-10,
                        "n": 1.0078400148329645,
                        "R_s_Ohm": 0.10336631172987705,
                        "G_p_S": 0.0,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.985176856856757,
                        "I_rs_A": 3.776521375198612e-10,
                        "n": 0.9979492671100877,
                        "R_s_Ohm": 0.1046256655202169,
                        "G_p_S": 0.000143503776517064,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.008357526744263,
                        "I_rs_A": 4.216692368218051e-10,
                        "n": 1.0028765413600613,
                        "R_s_Ohm": 0.10185559611933805,
                        "G_p_S": 0.0007737382436657485,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.001888427194626,
                        "I_rs_A": 1.0672169187595873e-09,
                        "n": 1.0433627261513425,
                        "R_s_Ohm": 0.08506456150121589,
                        "G_p_S": 0.0003084742981598182,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 7.994503102787226,
                        "I_rs_A": 3.9062717967302944e-10,
                        "n": 0.9994025123889798,
                        "R_s_Ohm": 0.10414753204745408,
                        "G_p_S": 0.00027008055800004007,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 8.006341540267963,
                        "I_rs_A": 3.1884234210989573e-10,
                        "n": 0.9912545243477779,
                        "R_s_Ohm": 0.10986159206082621,
                        "G_p_S": 0.0003617097086892112,
                    },
                ],
            },
        },
        {
            "given": {
                "test_set": "case3b",
                "rtol": 1e-05,
                "atol": 1e-08,
            },
            "expected": {
                "model_parameters": [
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9994666287062095,
                        "I_rs_A": 2.9354651943583707e-08,
                        "n": 1.2981616413345773,
                        "R_s_Ohm": 1.0168433580893808,
                        "G_p_S": 0.003301140718924285,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.000877280406033,
                        "I_rs_A": 3.789469293928319e-08,
                        "n": 1.3176755641831606,
                        "R_s_Ohm": 0.9128235334972428,
                        "G_p_S": 0.0033569790506572748,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0000593170791112,
                        "I_rs_A": 5.4857106962717544e-08,
                        "n": 1.3464352870042025,
                        "R_s_Ohm": 0.7775782576396228,
                        "G_p_S": 0.00333290664259892,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9995180566049039,
                        "I_rs_A": 2.231975666254148e-08,
                        "n": 1.2785138544377035,
                        "R_s_Ohm": 1.116494552637509,
                        "G_p_S": 0.0033122072535487416,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9982073207203306,
                        "I_rs_A": 4.073624431832634e-08,
                        "n": 1.3229785914778909,
                        "R_s_Ohm": 0.9199898308073255,
                        "G_p_S": 0.0032248478636654793,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0000391361345267,
                        "I_rs_A": 2.2019544091558122e-08,
                        "n": 1.277527816454505,
                        "R_s_Ohm": 1.112536008332753,
                        "G_p_S": 0.0033443642720931404,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9995962052685531,
                        "I_rs_A": 2.2624446596829243e-08,
                        "n": 1.2790323074518715,
                        "R_s_Ohm": 1.0730822886507685,
                        "G_p_S": 0.003329863121506824,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0011908375185106,
                        "I_rs_A": 2.7710204574767686e-08,
                        "n": 1.2940094476165043,
                        "R_s_Ohm": 1.0022632890559584,
                        "G_p_S": 0.0033933586545035577,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9992392644116262,
                        "I_rs_A": 3.660574622722916e-08,
                        "n": 1.3145937128881997,
                        "R_s_Ohm": 0.919105463033134,
                        "G_p_S": 0.00331077666767827,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.998673818535462,
                        "I_rs_A": 1.791696959899589e-08,
                        "n": 1.2623116127441572,
                        "R_s_Ohm": 1.1496470552187406,
                        "G_p_S": 0.0033066053187354332,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0005785246285626,
                        "I_rs_A": 2.1376151299115905e-08,
                        "n": 1.275141192287173,
                        "R_s_Ohm": 1.092562078471095,
                        "G_p_S": 0.003361562015494026,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.000802242873265,
                        "I_rs_A": 1.9262848501028223e-08,
                        "n": 1.2679950183439437,
                        "R_s_Ohm": 1.1334693713555342,
                        "G_p_S": 0.0034129876629129963,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0000189109891762,
                        "I_rs_A": 5.214615810633445e-08,
                        "n": 1.3427759837505093,
                        "R_s_Ohm": 0.8149862814841636,
                        "G_p_S": 0.0033073551780607724,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9996113687014543,
                        "I_rs_A": 4.969951464278507e-08,
                        "n": 1.3386046001123013,
                        "R_s_Ohm": 0.8420742244180227,
                        "G_p_S": 0.0032672256889604745,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9975053573232044,
                        "I_rs_A": 4.4084975522328236e-08,
                        "n": 1.32903188276204,
                        "R_s_Ohm": 0.8612440155668096,
                        "G_p_S": 0.0032712793933889683,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0021520160517008,
                        "I_rs_A": 2.4285768606410903e-08,
                        "n": 1.284573235638036,
                        "R_s_Ohm": 1.0775372908124643,
                        "G_p_S": 0.00339516986412654,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.000204933803175,
                        "I_rs_A": 3.2761285752028825e-08,
                        "n": 1.3062916572495027,
                        "R_s_Ohm": 0.9453271852392171,
                        "G_p_S": 0.00336162552129351,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.998846472938637,
                        "I_rs_A": 5.0216487926595656e-08,
                        "n": 1.3395537641900441,
                        "R_s_Ohm": 0.8099646249074971,
                        "G_p_S": 0.003271037155736449,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9983035003830415,
                        "I_rs_A": 3.30969776774477e-08,
                        "n": 1.3071002319974687,
                        "R_s_Ohm": 0.9422327351457126,
                        "G_p_S": 0.003311128738908875,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9996260645519497,
                        "I_rs_A": 3.980438126437782e-08,
                        "n": 1.321444705856373,
                        "R_s_Ohm": 0.9203002572848953,
                        "G_p_S": 0.0033144359883653515,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0021950927131764,
                        "I_rs_A": 1.537591317485395e-08,
                        "n": 1.2520492568572967,
                        "R_s_Ohm": 1.1977391283851322,
                        "G_p_S": 0.0034442929866516274,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.001506577581387,
                        "I_rs_A": 2.6926132720133457e-08,
                        "n": 1.2920656343310801,
                        "R_s_Ohm": 1.0072492394036772,
                        "G_p_S": 0.003424970169422707,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0007872904453994,
                        "I_rs_A": 2.21673197260315e-08,
                        "n": 1.2778576758023943,
                        "R_s_Ohm": 1.0834453395368246,
                        "G_p_S": 0.0034202850521235194,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0024016059061074,
                        "I_rs_A": 2.383004848008136e-08,
                        "n": 1.2829941318712303,
                        "R_s_Ohm": 1.0523684333914167,
                        "G_p_S": 0.003410350994191969,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9995074160130835,
                        "I_rs_A": 2.1072904994144047e-08,
                        "n": 1.274117685659872,
                        "R_s_Ohm": 1.1167085875111262,
                        "G_p_S": 0.0033121892555806496,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0003969889961142,
                        "I_rs_A": 2.7602746086632238e-08,
                        "n": 1.2936868969895046,
                        "R_s_Ohm": 1.0363268977297864,
                        "G_p_S": 0.003306029278364203,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0005080747286255,
                        "I_rs_A": 2.3730880201142218e-08,
                        "n": 1.2823902431754255,
                        "R_s_Ohm": 1.0421727761110142,
                        "G_p_S": 0.0033740231494503387,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9986861882049981,
                        "I_rs_A": 1.704855875267222e-08,
                        "n": 1.2591612419513054,
                        "R_s_Ohm": 1.1333271873065658,
                        "G_p_S": 0.003391732274059249,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0005916538048227,
                        "I_rs_A": 2.7127687166033234e-08,
                        "n": 1.2926696803803657,
                        "R_s_Ohm": 1.0393670328321023,
                        "G_p_S": 0.0033352295401007544,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0016336651697422,
                        "I_rs_A": 3.843531805753328e-08,
                        "n": 1.3187466771564087,
                        "R_s_Ohm": 0.9125486867557677,
                        "G_p_S": 0.0033641845809463783,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.000747763854674,
                        "I_rs_A": 1.5693955052937998e-08,
                        "n": 1.2535756011607004,
                        "R_s_Ohm": 1.1834444730802567,
                        "G_p_S": 0.003438810044790392,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.000229825853156,
                        "I_rs_A": 1.6423592307330904e-08,
                        "n": 1.256486457919137,
                        "R_s_Ohm": 1.1576133014873755,
                        "G_p_S": 0.0034227194640717457,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9973996150390215,
                        "I_rs_A": 3.423942834963075e-08,
                        "n": 1.3098247772849498,
                        "R_s_Ohm": 0.9637607483176308,
                        "G_p_S": 0.0032285985789927397,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.998319391967308,
                        "I_rs_A": 3.7901866080981426e-08,
                        "n": 1.3175104397738175,
                        "R_s_Ohm": 0.9071663629666117,
                        "G_p_S": 0.0032756665372975073,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9988004667154763,
                        "I_rs_A": 3.1973131093190846e-08,
                        "n": 1.3049270496906,
                        "R_s_Ohm": 0.9691068711591652,
                        "G_p_S": 0.00329382270107873,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9997888888435099,
                        "I_rs_A": 3.107469683012977e-08,
                        "n": 1.30270069620285,
                        "R_s_Ohm": 0.9714442365923804,
                        "G_p_S": 0.00335353486505934,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0011693966205655,
                        "I_rs_A": 2.2064872213780453e-08,
                        "n": 1.2770680351610493,
                        "R_s_Ohm": 1.0693252145169938,
                        "G_p_S": 0.003409102366972087,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0005178412288815,
                        "I_rs_A": 2.548933524022367e-08,
                        "n": 1.2874799738366551,
                        "R_s_Ohm": 1.050439428378089,
                        "G_p_S": 0.003319376359328255,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0016599999200073,
                        "I_rs_A": 2.427019057219945e-08,
                        "n": 1.284558942589953,
                        "R_s_Ohm": 1.0691525786327134,
                        "G_p_S": 0.003378665253804484,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9997771481079374,
                        "I_rs_A": 4.34376748868677e-08,
                        "n": 1.3281684060245798,
                        "R_s_Ohm": 0.8885539429897849,
                        "G_p_S": 0.0033133275505685656,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9972487438422752,
                        "I_rs_A": 7.208591008732503e-08,
                        "n": 1.3687513045729216,
                        "R_s_Ohm": 0.7265319938152016,
                        "G_p_S": 0.003221681260621011,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9997415321640246,
                        "I_rs_A": 2.9658851648124743e-08,
                        "n": 1.299217618681266,
                        "R_s_Ohm": 0.998831453906894,
                        "G_p_S": 0.0033311571214275416,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9999037627151482,
                        "I_rs_A": 2.839178702407927e-08,
                        "n": 1.2956083809847851,
                        "R_s_Ohm": 0.9988135994955747,
                        "G_p_S": 0.0033732424337032108,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0014758074077226,
                        "I_rs_A": 3.121329835168594e-08,
                        "n": 1.30304771493829,
                        "R_s_Ohm": 0.9764721278262312,
                        "G_p_S": 0.0034135637775218075,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0014576450898243,
                        "I_rs_A": 3.4228083271947694e-08,
                        "n": 1.3100134515168798,
                        "R_s_Ohm": 0.9549334589408411,
                        "G_p_S": 0.0033757107890388205,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0005258567821245,
                        "I_rs_A": 3.522989509160631e-08,
                        "n": 1.3124418399782132,
                        "R_s_Ohm": 0.9875745612267234,
                        "G_p_S": 0.0033371225215826526,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0001029292792951,
                        "I_rs_A": 3.3100739523234346e-08,
                        "n": 1.3075605960712358,
                        "R_s_Ohm": 0.9600332365882359,
                        "G_p_S": 0.0033317920371938765,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9974361792151238,
                        "I_rs_A": 4.945537406545297e-08,
                        "n": 1.3386181361386125,
                        "R_s_Ohm": 0.8359161309444492,
                        "G_p_S": 0.0032480561417978526,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 0.9998106171711685,
                        "I_rs_A": 2.7715589103406788e-08,
                        "n": 1.2942106505127018,
                        "R_s_Ohm": 1.0461205372548008,
                        "G_p_S": 0.003297740326858018,
                    },
                    {
                        "N_s": 72,
                        "T_degC": 25.0,
                        "I_ph_A": 1.0011575920320366,
                        "I_rs_A": 4.191224263256794e-08,
                        "n": 1.325402076402849,
                        "R_s_Ohm": 0.8463502652642775,
                        "G_p_S": 0.0034149338161497747,
                    },
                ],
            },
        },
        {
            "given": {
                "test_set": "case3c",
                "rtol": 1e-05,
                "atol": 1e-08,
            },
            "expected": {
                "model_parameters": [
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4997369516623484,
                        "I_rs_A": 1.5061118161802645e-09,
                        "n": 1.3243915674486035,
                        "R_s_Ohm": 0.005335207899449487,
                        "G_p_S": 0.0002867840070292473,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.506044535804447,
                        "I_rs_A": 8.577705852983398e-10,
                        "n": 1.2907013492175128,
                        "R_s_Ohm": 0.1101716846553498,
                        "G_p_S": 0.0004284449310800067,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.502249405095908,
                        "I_rs_A": 1.2252014672722328e-09,
                        "n": 1.312134098942711,
                        "R_s_Ohm": 0.07480569759434438,
                        "G_p_S": 0.0003140961523714588,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4992241860073166,
                        "I_rs_A": 7.24102509375917e-10,
                        "n": 1.2807354906678423,
                        "R_s_Ohm": 0.13431202013560037,
                        "G_p_S": 0.00037533239561293155,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.496561084398025,
                        "I_rs_A": 9.8347828027294e-10,
                        "n": 1.2996131983879926,
                        "R_s_Ohm": 0.1273210766297349,
                        "G_p_S": 0.0002939863469928075,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.500443719305679,
                        "I_rs_A": 1.1113616830423715e-09,
                        "n": 1.3067694477169196,
                        "R_s_Ohm": 0.07449191611307941,
                        "G_p_S": 0.0003778868275076456,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4979649542993845,
                        "I_rs_A": 1.450429718615067e-09,
                        "n": 1.3226563356051992,
                        "R_s_Ohm": 0.041084803429213605,
                        "G_p_S": 0.0002852853165043752,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4997674497739037,
                        "I_rs_A": 8.620908159642268e-10,
                        "n": 1.2911430663485923,
                        "R_s_Ohm": 0.10213901401031662,
                        "G_p_S": 0.00036047786155585877,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5036735034066155,
                        "I_rs_A": 1.0653896893724758e-09,
                        "n": 1.3037786481158757,
                        "R_s_Ohm": 0.089496503409356,
                        "G_p_S": 0.00033126618348088113,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.503205997714418,
                        "I_rs_A": 7.152984916758761e-10,
                        "n": 1.2807214826521756,
                        "R_s_Ohm": 0.14430889205546218,
                        "G_p_S": 0.0004414064870092915,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.502588238607673,
                        "I_rs_A": 8.786794152082726e-10,
                        "n": 1.2918446259800085,
                        "R_s_Ohm": 0.09731458818044524,
                        "G_p_S": 0.0003376663325243596,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4996941668927852,
                        "I_rs_A": 1.0957142498406727e-09,
                        "n": 1.3053539808627481,
                        "R_s_Ohm": 0.09043615930021061,
                        "G_p_S": 0.0002938464313800848,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.497920163070206,
                        "I_rs_A": 1.1273784909158006e-09,
                        "n": 1.3070549059019485,
                        "R_s_Ohm": 0.08720427707591634,
                        "G_p_S": 0.0002660189906489103,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4986114643635826,
                        "I_rs_A": 1.5985225476221278e-09,
                        "n": 1.328436097883809,
                        "R_s_Ohm": 0.0,
                        "G_p_S": 0.0003079658223835598,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5010143567536947,
                        "I_rs_A": 1.8817423531722026e-09,
                        "n": 1.3387542677795685,
                        "R_s_Ohm": 0.0,
                        "G_p_S": 0.0002661142704760889,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4994262343107803,
                        "I_rs_A": 7.984658062502975e-10,
                        "n": 1.286781095675608,
                        "R_s_Ohm": 0.15129164019007096,
                        "G_p_S": 0.0003461439051425951,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4976869800575843,
                        "I_rs_A": 8.737399329037445e-10,
                        "n": 1.292255241983129,
                        "R_s_Ohm": 0.13011344655777027,
                        "G_p_S": 0.0003180191633419827,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.50619380092642,
                        "I_rs_A": 1.1354563742499066e-09,
                        "n": 1.3077473738659282,
                        "R_s_Ohm": 0.07614498153741181,
                        "G_p_S": 0.00037471128906362195,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.504667541629555,
                        "I_rs_A": 7.722392589716563e-10,
                        "n": 1.2845882322758917,
                        "R_s_Ohm": 0.12797351722890327,
                        "G_p_S": 0.00041718646739971176,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4972422305890265,
                        "I_rs_A": 1.0581556856829705e-09,
                        "n": 1.3033358535904604,
                        "R_s_Ohm": 0.10063162492211673,
                        "G_p_S": 0.00028129149515640717,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.502750571022787,
                        "I_rs_A": 4.444732984572404e-10,
                        "n": 1.253437369791456,
                        "R_s_Ohm": 0.2349659573544369,
                        "G_p_S": 0.0004293716240484672,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5004244610760353,
                        "I_rs_A": 7.17428333411032e-10,
                        "n": 1.2807128859576011,
                        "R_s_Ohm": 0.1438168083250685,
                        "G_p_S": 0.0004146934917188981,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.498136249573161,
                        "I_rs_A": 6.645757524689046e-10,
                        "n": 1.2758082756849827,
                        "R_s_Ohm": 0.15357519799311453,
                        "G_p_S": 0.00031470764149488044,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4982872999061696,
                        "I_rs_A": 9.123556512906233e-10,
                        "n": 1.2946804382086037,
                        "R_s_Ohm": 0.12649465053534698,
                        "G_p_S": 0.00031669829820271016,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4952919308379133,
                        "I_rs_A": 9.341958947289223e-10,
                        "n": 1.2962443795176855,
                        "R_s_Ohm": 0.1310984554284669,
                        "G_p_S": 0.0002968824066164476,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.49739983544053,
                        "I_rs_A": 7.783268616911325e-10,
                        "n": 1.2850615979785767,
                        "R_s_Ohm": 0.15223495262955697,
                        "G_p_S": 0.00025884412031640543,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4962198960760658,
                        "I_rs_A": 1.1882161941360648e-09,
                        "n": 1.310247762096832,
                        "R_s_Ohm": 0.08522984780565043,
                        "G_p_S": 0.00030999667037123177,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.498724036717211,
                        "I_rs_A": 7.581468492116968e-10,
                        "n": 1.2835171119028321,
                        "R_s_Ohm": 0.12468225004835502,
                        "G_p_S": 0.0003438419299687208,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4968486772628316,
                        "I_rs_A": 8.653307383440024e-10,
                        "n": 1.2916958941145813,
                        "R_s_Ohm": 0.1505174809338452,
                        "G_p_S": 0.00028852276511811235,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5030472488344047,
                        "I_rs_A": 6.672323152687148e-10,
                        "n": 1.2763010636917032,
                        "R_s_Ohm": 0.17052305626114503,
                        "G_p_S": 0.00033421548960116,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5006189028416346,
                        "I_rs_A": 1.6136277126481336e-09,
                        "n": 1.3287012604427957,
                        "R_s_Ohm": 0.0,
                        "G_p_S": 0.0003191755491013179,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.504438172408033,
                        "I_rs_A": 3.6466605524585234e-10,
                        "n": 1.2431951457150845,
                        "R_s_Ohm": 0.2995812377011448,
                        "G_p_S": 0.00042954989290359843,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.498200473422625,
                        "I_rs_A": 1.16006704888709e-09,
                        "n": 1.308652301488261,
                        "R_s_Ohm": 0.0353275265572867,
                        "G_p_S": 0.00031928600296305055,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.50081780778657,
                        "I_rs_A": 1.7426675424997309e-09,
                        "n": 1.3341802039178665,
                        "R_s_Ohm": 0.016123006437771793,
                        "G_p_S": 0.00031515962354327444,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.498131438658181,
                        "I_rs_A": 1.239097595640563e-09,
                        "n": 1.3130955925185754,
                        "R_s_Ohm": 0.06567687957386366,
                        "G_p_S": 0.0003001657230199403,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4992533679010287,
                        "I_rs_A": 6.472221295212729e-10,
                        "n": 1.2749674782687936,
                        "R_s_Ohm": 0.20210423664867308,
                        "G_p_S": 0.00032839952867905577,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.500002176231359,
                        "I_rs_A": 1.3284238108213304e-09,
                        "n": 1.3170748190497321,
                        "R_s_Ohm": 0.0655180952848773,
                        "G_p_S": 0.0003038087903256503,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5028653669384426,
                        "I_rs_A": 8.340328042970445e-10,
                        "n": 1.2890295953366784,
                        "R_s_Ohm": 0.12117801313096273,
                        "G_p_S": 0.0003492821080175468,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.503807764139565,
                        "I_rs_A": 1.3271412820042015e-09,
                        "n": 1.3167771709818923,
                        "R_s_Ohm": 0.029965728963499156,
                        "G_p_S": 0.0003946863323266931,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.499621314068866,
                        "I_rs_A": 6.180898997528074e-10,
                        "n": 1.271504940166591,
                        "R_s_Ohm": 0.15086479392061417,
                        "G_p_S": 0.0003333419049876781,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4983754430340093,
                        "I_rs_A": 1.2856427855483316e-09,
                        "n": 1.3155326505617209,
                        "R_s_Ohm": 0.06384696054940922,
                        "G_p_S": 0.0003426393923426863,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.504312526935498,
                        "I_rs_A": 9.16276597862128e-10,
                        "n": 1.2948338747287875,
                        "R_s_Ohm": 0.10687455831056654,
                        "G_p_S": 0.0003774389396401772,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5009115149969916,
                        "I_rs_A": 1.8852593737052563e-09,
                        "n": 1.33931796327696,
                        "R_s_Ohm": 0.0,
                        "G_p_S": 0.00033083314576667856,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5003880606574125,
                        "I_rs_A": 1.0446873341261143e-09,
                        "n": 1.302591845692005,
                        "R_s_Ohm": 0.1100513715335152,
                        "G_p_S": 0.00032182187064869227,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5017130939582,
                        "I_rs_A": 1.718621044409048e-09,
                        "n": 1.3330766292455185,
                        "R_s_Ohm": 0.0,
                        "G_p_S": 0.0003476765937267845,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.501092128248128,
                        "I_rs_A": 7.553569757907968e-10,
                        "n": 1.2839186435587624,
                        "R_s_Ohm": 0.18077342281876319,
                        "G_p_S": 0.00030307996070794125,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.5002097131815213,
                        "I_rs_A": 5.905539325940397e-10,
                        "n": 1.2695173198418348,
                        "R_s_Ohm": 0.20214099818115186,
                        "G_p_S": 0.0003428545105206218,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.4992878898636888,
                        "I_rs_A": 8.215663772491108e-10,
                        "n": 1.2884684469973533,
                        "R_s_Ohm": 0.13481632667396362,
                        "G_p_S": 0.00030799377421943764,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.500561669979768,
                        "I_rs_A": 1.1448441812657025e-09,
                        "n": 1.3074352606184865,
                        "R_s_Ohm": 0.04255794563216341,
                        "G_p_S": 0.00036374220610309077,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 2.502150299784716,
                        "I_rs_A": 1.2295021205902282e-09,
                        "n": 1.3120612556426265,
                        "R_s_Ohm": 0.0351686870644355,
                        "G_p_S": 0.0003457457048231443,
                    },
                ],
            },
        },
        {
            "given": {
                "test_set": "case3d",
                "rtol": 0.0005,
                "atol": 0.004,
            },
            "expected": {
                "model_parameters": [
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.4985389563141154,
                        "I_rs_A": 1.4124850896913641e-08,
                        "n": 1.5315901039236726,
                        "R_s_Ohm": 0.1367787219482562,
                        "G_p_S": 0.0033220548157582366,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5010609876015714,
                        "I_rs_A": 9.461160797079433e-09,
                        "n": 1.4954332917475195,
                        "R_s_Ohm": 1.0790103160989297,
                        "G_p_S": 0.003349225102099971,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.4992295870377278,
                        "I_rs_A": 1.0244107432143395e-08,
                        "n": 1.501929339203371,
                        "R_s_Ohm": 0.8556670274906096,
                        "G_p_S": 0.0033247240232029507,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49923733939208037,
                        "I_rs_A": 1.0123479213395354e-08,
                        "n": 1.5007290268336437,
                        "R_s_Ohm": 0.922212776795974,
                        "G_p_S": 0.0033209175239039936,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49912737854904965,
                        "I_rs_A": 1.204300398970007e-08,
                        "n": 1.516648185975785,
                        "R_s_Ohm": 0.8455318128752964,
                        "G_p_S": 0.0033155502862121863,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.499877166915085,
                        "I_rs_A": 1.3097403622047231e-08,
                        "n": 1.5244866786104259,
                        "R_s_Ohm": 0.33205664182730205,
                        "G_p_S": 0.0033334184222250855,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49873444547665935,
                        "I_rs_A": 1.3255802869833801e-08,
                        "n": 1.5260572957806395,
                        "R_s_Ohm": 0.38841400361369666,
                        "G_p_S": 0.003322883107176686,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5006209786535393,
                        "I_rs_A": 7.948310693170435e-09,
                        "n": 1.4805745910457482,
                        "R_s_Ohm": 1.4441444555351513,
                        "G_p_S": 0.0033464318449142596,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49802193783580123,
                        "I_rs_A": 1.769486922854351e-08,
                        "n": 1.5527775649209834,
                        "R_s_Ohm": 0.0,
                        "G_p_S": 0.0033074697867400134,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49791251064944486,
                        "I_rs_A": 1.6117620538675232e-08,
                        "n": 1.5438209279581756,
                        "R_s_Ohm": 0.0886403313786087,
                        "G_p_S": 0.003307740069767556,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5011515446975588,
                        "I_rs_A": 7.761949606341384e-09,
                        "n": 1.478053752866171,
                        "R_s_Ohm": 1.4884131593497287,
                        "G_p_S": 0.0033490402538067436,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49874242716896466,
                        "I_rs_A": 1.2193991858941495e-08,
                        "n": 1.5181263324814414,
                        "R_s_Ohm": 0.176087266877603,
                        "G_p_S": 0.0033332517729494776,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5026831773930577,
                        "I_rs_A": 6.857387701514742e-09,
                        "n": 1.4674478217501195,
                        "R_s_Ohm": 2.0949659183182665,
                        "G_p_S": 0.0033602399182836693,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.4983634457075293,
                        "I_rs_A": 1.5041545514073948e-08,
                        "n": 1.5371868724512414,
                        "R_s_Ohm": 0.0,
                        "G_p_S": 0.0033159186615528496,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5021874320083239,
                        "I_rs_A": 6.971363338932634e-09,
                        "n": 1.4683793302282109,
                        "R_s_Ohm": 1.714511455454418,
                        "G_p_S": 0.003356945064212616,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5021795213008358,
                        "I_rs_A": 5.1077619494783885e-09,
                        "n": 1.4424841299031947,
                        "R_s_Ohm": 2.3946793895603045,
                        "G_p_S": 0.0033545205293497586,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49900815244125496,
                        "I_rs_A": 1.2097077047334011e-08,
                        "n": 1.5172058878907282,
                        "R_s_Ohm": 0.4556210222966506,
                        "G_p_S": 0.0033264290315723473,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5013834192750101,
                        "I_rs_A": 6.444616494861552e-09,
                        "n": 1.4615284593840858,
                        "R_s_Ohm": 1.89188444906294,
                        "G_p_S": 0.00335139612645882,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5005802304989647,
                        "I_rs_A": 9.370653954818161e-09,
                        "n": 1.4944826777878026,
                        "R_s_Ohm": 1.3313812438177497,
                        "G_p_S": 0.0033384965403251837,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5042455972056147,
                        "I_rs_A": 3.735679692091375e-09,
                        "n": 1.417549074347812,
                        "R_s_Ohm": 3.382161039841245,
                        "G_p_S": 0.003372611870138592,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5037654785439782,
                        "I_rs_A": 4.693483774021149e-09,
                        "n": 1.4359592160170043,
                        "R_s_Ohm": 2.6646551108368612,
                        "G_p_S": 0.00337337487413985,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5017136974397864,
                        "I_rs_A": 7.180452207382916e-09,
                        "n": 1.4709875949948712,
                        "R_s_Ohm": 1.875524390432754,
                        "G_p_S": 0.003343201319053123,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5012924931549472,
                        "I_rs_A": 6.4815758106308386e-09,
                        "n": 1.462355962562205,
                        "R_s_Ohm": 1.7773789359098364,
                        "G_p_S": 0.0033497374148665376,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5030853727318901,
                        "I_rs_A": 5.625125751845028e-09,
                        "n": 1.4509524557370066,
                        "R_s_Ohm": 2.3458420969405864,
                        "G_p_S": 0.0033708289075647813,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5026048513101937,
                        "I_rs_A": 6.936157542446472e-09,
                        "n": 1.4683585793630562,
                        "R_s_Ohm": 1.9694915134930178,
                        "G_p_S": 0.00336167341165928,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5002861945335847,
                        "I_rs_A": 1.0217066456926013e-08,
                        "n": 1.501958017022914,
                        "R_s_Ohm": 1.0134450824031707,
                        "G_p_S": 0.00333541069929178,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49876653569615087,
                        "I_rs_A": 1.2123589717140855e-08,
                        "n": 1.5175671650412872,
                        "R_s_Ohm": 0.40906957558129586,
                        "G_p_S": 0.003323610661337325,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.4987198536761773,
                        "I_rs_A": 1.2478379380906408e-08,
                        "n": 1.519989939421951,
                        "R_s_Ohm": 0.48855287866675423,
                        "G_p_S": 0.0033216376570643536,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5007426603209458,
                        "I_rs_A": 9.146809836357468e-09,
                        "n": 1.4926679987171008,
                        "R_s_Ohm": 1.4467430732282704,
                        "G_p_S": 0.003340028346525834,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5008951988381921,
                        "I_rs_A": 6.806079623160219e-09,
                        "n": 1.4665715308957195,
                        "R_s_Ohm": 1.9208212319378277,
                        "G_p_S": 0.0033369523604813924,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49807808616359966,
                        "I_rs_A": 1.478954040953276e-08,
                        "n": 1.5355671626775371,
                        "R_s_Ohm": 0.0,
                        "G_p_S": 0.0033172987916865793,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5034719824126963,
                        "I_rs_A": 4.541805072278035e-09,
                        "n": 1.4335095058685376,
                        "R_s_Ohm": 2.9369938246851506,
                        "G_p_S": 0.003370127783896666,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5035803423539333,
                        "I_rs_A": 5.248881420630926e-09,
                        "n": 1.4452216236720927,
                        "R_s_Ohm": 2.8789810773901547,
                        "G_p_S": 0.0033662543326308278,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.50043631973556,
                        "I_rs_A": 8.150315076804641e-09,
                        "n": 1.4822205555117862,
                        "R_s_Ohm": 1.5157050883480152,
                        "G_p_S": 0.0033356859997321795,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49848963301586274,
                        "I_rs_A": 1.4944128795384666e-08,
                        "n": 1.5362597605654145,
                        "R_s_Ohm": 0.0,
                        "G_p_S": 0.0033165416000465278,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.4985166734420743,
                        "I_rs_A": 1.4519671043856778e-08,
                        "n": 1.5338232415881976,
                        "R_s_Ohm": 0.10853845522974984,
                        "G_p_S": 0.0033224443170507753,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5013109390195483,
                        "I_rs_A": 7.486688180970608e-09,
                        "n": 1.4749012812051383,
                        "R_s_Ohm": 2.0046551108142343,
                        "G_p_S": 0.0033425478954245496,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5005421128562884,
                        "I_rs_A": 6.5508473869937996e-09,
                        "n": 1.4630668248814866,
                        "R_s_Ohm": 1.8565844086947494,
                        "G_p_S": 0.003342113913525929,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.4989812842228938,
                        "I_rs_A": 1.3287660493682857e-08,
                        "n": 1.5253572413524197,
                        "R_s_Ohm": 0.043720548627888364,
                        "G_p_S": 0.0033255174732873772,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49942800098704826,
                        "I_rs_A": 1.2091224187991032e-08,
                        "n": 1.5168357877033063,
                        "R_s_Ohm": 0.5166250564529055,
                        "G_p_S": 0.0033271519245014042,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5006529009087197,
                        "I_rs_A": 8.015338168984404e-09,
                        "n": 1.4803584879351195,
                        "R_s_Ohm": 1.6030178145912948,
                        "G_p_S": 0.003334721518269878,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49987835440511996,
                        "I_rs_A": 1.1222892353541266e-08,
                        "n": 1.5105545014258768,
                        "R_s_Ohm": 0.8236125610893129,
                        "G_p_S": 0.0033330853816245553,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5036789122954035,
                        "I_rs_A": 4.292726549352086e-09,
                        "n": 1.428769782817743,
                        "R_s_Ohm": 3.2409021242622567,
                        "G_p_S": 0.0033632400191537337,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.4988912236371362,
                        "I_rs_A": 1.4156118139549834e-08,
                        "n": 1.5318510372673078,
                        "R_s_Ohm": 0.402761183853545,
                        "G_p_S": 0.0033209135526656827,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49899178284688434,
                        "I_rs_A": 1.5217645678585322e-08,
                        "n": 1.5384736322061767,
                        "R_s_Ohm": 0.0,
                        "G_p_S": 0.003325279747936424,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.5013224078492119,
                        "I_rs_A": 7.4236555086499925e-09,
                        "n": 1.474158330351258,
                        "R_s_Ohm": 1.6329936236212856,
                        "G_p_S": 0.0033487513948198236,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.4993018367603058,
                        "I_rs_A": 1.0723984468043446e-08,
                        "n": 1.506877683994314,
                        "R_s_Ohm": 0.7899349255772301,
                        "G_p_S": 0.0033338158162380376,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49835328875941787,
                        "I_rs_A": 1.6567982796223286e-08,
                        "n": 1.5466529051503135,
                        "R_s_Ohm": 0.0,
                        "G_p_S": 0.0033161611057395397,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49922192562852824,
                        "I_rs_A": 1.1212949304098514e-08,
                        "n": 1.5102489456369623,
                        "R_s_Ohm": 0.5234075815834526,
                        "G_p_S": 0.0033332609788524936,
                    },
                    {
                        "N_s": 140,
                        "T_degC": 25.0,
                        "I_ph_A": 0.49852894514677704,
                        "I_rs_A": 1.5516611280532036e-08,
                        "n": 1.5402198844299178,
                        "R_s_Ohm": 0.0,
                        "G_p_S": 0.0033204909663765115,
                    },
                ],
            },
        },
    ],
)
def fit_fixture(request):
    return request.param


def test_fit(fit_fixture):
    """Test SDE fit against standard curves provided by ivcurves package."""
    given = fit_fixture["given"]
    expected = fit_fixture["expected"]

    base_path = IVCURVES_MULTIPLEXED_PATH.joinpath("test_sets")
    test_set = given["test_set"]

    with base_path.joinpath(test_set + ".json").open(encoding="utf8") as file:
        iv_curve_json = json.load(file)

    with importlib.resources.as_file(base_path.joinpath(test_set + ".csv")) as file:
        model_parameters_true_df = pandas.read_csv(file, encoding="utf-8")

    model_parameters_got_df = pandas.DataFrame(
        columns=[
            "Index",
            "I_ph_A",
            "I_rs_A",
            "n",
            "R_s_Ohm",
            "G_p_S",
            "N_s",
            "T_degC",
        ]
    )

    N_s = iv_curve_json["cells_in_series"]

    for idx, iv_curve_set in enumerate(iv_curve_json["IV Curves"]):
        model_parameters_expected = expected["model_parameters"][idx]

        Index = iv_curve_set["Index"]
        T_degC = scipy.constants.convert_temperature(
            float(iv_curve_set["Temperature"]), "Kelvin", "Celsius"
        )
        V_V = numpy.array([float(voltage) for voltage in iv_curve_set["Voltages"]])
        I_A = numpy.array([float(current) for current in iv_curve_set["Currents"]])

        # Typical fit where all ICs derived from I-V curve and I-V curve parameters are
        # provided for scaling.
        iv_curve = IVCurve(V_V=V_V, I_A=I_A)
        model_parameters_unfittable = ModelParametersUnfittable(N_s=N_s, T_degC=T_degC)

        model_parameters_got = inference_iv_curve.fit(
            iv_curve=iv_curve,
            model_parameters_unfittable=model_parameters_unfittable,
        )["model_parameters"]

        for key in model_parameters_got:
            numpy.testing.assert_allclose(
                model_parameters_got[key],
                model_parameters_expected[key],
                rtol=given["rtol"],
                atol=given["atol"],
                err_msg=f"got {model_parameters_got[key]} for {key}, expected {model_parameters_expected[key]}",
            )

        model_parameters_got_df.loc[idx, "Index"] = Index
        model_parameters_got_df.loc[idx, "I_ph_A"] = model_parameters_got["I_ph_A"]
        model_parameters_got_df.loc[idx, "I_rs_A"] = model_parameters_got["I_rs_A"]
        model_parameters_got_df.loc[idx, "n"] = model_parameters_got["n"]
        model_parameters_got_df.loc[idx, "R_s_Ohm"] = model_parameters_got["R_s_Ohm"]
        model_parameters_got_df.loc[idx, "G_p_S"] = model_parameters_got["G_p_S"]
        model_parameters_got_df.loc[idx, "N_s"] = model_parameters_got["N_s"]
        model_parameters_got_df.loc[idx, "T_degC"] = model_parameters_got["T_degC"]

        # Test using true parameters as ICs.
        if "case3" in test_set:
            # Only one true set of parameters.
            idx_true_ic = 0
        else:
            idx_true_ic = idx

        model_parameters_fittable_ic_provided_true = ModelParametersFittableProvided(
            I_ph_A=model_parameters_true_df.loc[idx_true_ic, "photocurrent"],
            I_rs_A=model_parameters_true_df.loc[idx_true_ic, "saturation_current"],
            n=model_parameters_true_df.loc[idx_true_ic, "n"],
            R_s_Ohm=model_parameters_true_df.loc[idx_true_ic, "resistance_series"],
            G_p_S=1 / model_parameters_true_df.loc[idx_true_ic, "resistance_shunt"],
        )

        # Ensure convergence when using true parameters for ICs.
        inference_iv_curve.fit(
            iv_curve=iv_curve,
            model_parameters_unfittable=model_parameters_unfittable,
            model_parameters_fittable_ic_provided=model_parameters_fittable_ic_provided_true,
        )

        # Ensure convergence when using true parameters for ICs with all fit parameters
        # fixed.
        inference_iv_curve.fit(
            iv_curve=iv_curve,
            model_parameters_unfittable=model_parameters_unfittable,
            model_parameters_fittable_ic_provided=model_parameters_fittable_ic_provided_true,
            model_parameters_fittable_fixed_provided=ModelParametersFittableFixedProvided(
                I_ph_A=True,
                I_rs_A=True,
                n=True,
                R_s_Ohm=True,
                G_p_S=True,
            ),
        )

    # This dataframe can be used to run benchmark from ivcurves.
    ivcurves_df = pandas.DataFrame(
        columns=[
            "Index",
            "photocurrent",
            "saturation_current",
            "resistance_series",
            "resistance_shunt",
            "n",
            "cells_in_series",
        ]
    )

    if "case3" in test_set:
        model_parameters_got_df_mean = model_parameters_got_df.mean(axis=0)
        ivcurves_df.loc[0, "Index"] = 1
        ivcurves_df.loc[0, "photocurrent"] = model_parameters_got_df_mean["I_ph_A"]
        ivcurves_df.loc[0, "saturation_current"] = model_parameters_got_df_mean[
            "I_rs_A"
        ]
        ivcurves_df.loc[0, "resistance_series"] = model_parameters_got_df_mean[
            "R_s_Ohm"
        ]
        ivcurves_df.loc[0, "resistance_shunt"] = (
            1 / model_parameters_got_df_mean["G_p_S"]
        )
        ivcurves_df.loc[0, "n"] = model_parameters_got_df_mean["n"]
        ivcurves_df.loc[0, "cells_in_series"] = N_s
    else:
        for df_idx, row in model_parameters_got_df.iterrows():
            ivcurves_df.loc[df_idx, "Index"] = row["Index"]
            ivcurves_df.loc[df_idx, "photocurrent"] = row["I_ph_A"]
            ivcurves_df.loc[df_idx, "saturation_current"] = row["I_rs_A"]
            ivcurves_df.loc[df_idx, "resistance_series"] = row["R_s_Ohm"]
            ivcurves_df.loc[df_idx, "resistance_shunt"] = 1 / row["G_p_S"]
            ivcurves_df.loc[df_idx, "n"] = row["n"]
            ivcurves_df.loc[df_idx, "cells_in_series"] = N_s

    # The file output here can be compared using ivcurves/compare_curves.py.
    ivcurves_df.to_csv(
        path_or_buf=TEST_ARTIFACTS_PATH.joinpath(test_set + ".csv"),
        index=False,
    )


# This must execute after all test_fit cases have finished.
def test_fit_benchmark():
    """Check ivcurves benchmark scores from test_fit tests."""
    with importlib.resources.as_file(
        IVCURVES_MULTIPLEXED_PATH.joinpath("compare_curves.py")
    ) as script:
        # This command creates CSV score files for each test-set case.
        return_value = os.system(
            f"python {script} {TEST_ARTIFACTS_PATH} --csv-output-path "
            f"{TEST_ARTIFACTS_PATH}"
        )

    # Check command success.
    assert return_value == 0

    # Read in created benchmark file.
    overall_scores_df_got = pandas.read_csv(
        TEST_ARTIFACTS_PATH.joinpath("overall_scores.csv"),
        index_col="test_set",
        encoding="utf-8",
    )

    overall_scores_got_test_set = set(overall_scores_df_got.index)
    overall_scores_expected_test_set = set(
        ("case1", "case2", "case3a", "case3b", "case3c", "case3d")
    )

    assert overall_scores_got_test_set == overall_scores_expected_test_set

    overall_scores_expected_score = {
        "case1": 30.0000000000000230624,
        "case2": 31.0000000000000336641,
        "case3a": 0.1068342249452320955,
        "case3b": 0.06836656715157551795,
        "case3c": 0.06659086185392485641,
        "case3d": 0.1756979161971756333,
    }

    for case in overall_scores_expected_score:
        numpy.testing.assert_allclose(
            overall_scores_df_got.loc[case, "score"],
            overall_scores_expected_score[case],
            rtol=5e-05,
            atol=8e-06,
        )
